<?php
/************************************************************************
 * Mailer v0.2.1-FINAL                                Start: 08/26/2003 *
 * ===================                          Last change: 11/29/2004 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : mysql-manager.php                                *
 * -------------------------------------------------------------------- *
 * Short description : All database-related functions                   *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Alle datenbank-relevanten Funktionen             *
 * -------------------------------------------------------------------- *
 * $Revision::                                                        $ *
 * $Date::                                                            $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author::                                                          $ *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * Copyright (c) 2009 - 2013 by Mailer Developer Team                   *
 * For more information visit: http://mxchange.org                      *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} // END - if

// "Getter" for module description
// @TODO Can we cache this?
function getTitleFromMenu ($mode, $what, $column = 'what', $ADD = '') {
	// Debug message
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'mode=' . $mode . ',what=' . $what . ',column=' . $column . ',add=' . $ADD);

	// Fix empty 'what'
	if (empty($what)) {
		$what = getIndexHome();
	} elseif ((isGetRequestElementSet('action')) && ($column == 'what')) {
		// Get it from action
		return getTitleFromMenu($mode, getAction(), 'action', $ADD);
	} elseif ($what == 'welcome') {
		// Overview page
		return '{--WHAT_IS_WELCOME--}';
	}

	// Default is not found
	$data['title'] = '??? (' . $what . ')';

	// Look for title
	$result = sqlQueryEscaped("SELECT `title` FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE `%s`='%s'" . $ADD . " LIMIT 1",
		array(
			$mode,
			$column,
			$what
		), __FUNCTION__, __LINE__);

	// Is there an entry?
	if (sqlNumRows($result) == 1) {
		// Fetch the title
		$data = sqlFetchArray($result);
	} // END - if

	// Free result
	sqlFreeResult($result);

	// Return it
	return $data['title'];
}

// Add link into output stream (or return it) for 'You Are Here' navigation
function addYouAreHereLink ($accessLevel, $FQFN, $return = FALSE) {
	// Use only filename of the FQFN...
	$file = basename($FQFN);

	// Init variables
	$linkAdd = '';
	$OUT = '';
	$ADD = '';
	$prefix = '';

	// First we have to do some analysis...
	if (substr($file, 0, 7) == 'action-') {
		// This is an action file!
		$type = 'action';
		$search = substr($file, 7);

		// Get access level from it
		$modCheck = getModuleFromFileName($file, $accessLevel);

		// Add what
		$ADD = " AND (`what`='' OR `what` IS NULL)";
	} elseif (substr($file, 0, 5) == 'what-') {
		// This is a 'what file'!
		$type = 'what';
		$search = substr($file, 5);

		// Get access level from it
		$modCheck = getModuleFromFileName($file, $accessLevel);

		// Is there admin? Then display all
		$ADD = " AND `visible`='Y' AND `locked`='N'";
		if (isAdmin()) {
			// Display all!
			$ADD = '';
		} // END - if

		$dummy = substr($search, 0, -4);
		$ADD .= sprintf(" AND `action`='%s'", getActionFromModuleWhat($accessLevel, $dummy));
	} elseif ($accessLevel == 'sponsor') {
		// Sponsor menu
		$type     = 'what';
		$search   = $file;
		$modCheck = getModule();
	} else {
		// Other
		$type     = 'menu';
		$search   = $file;
		$modCheck = getModule();
	}

	// Begin the navigation line
	if (!isset($GLOBALS['nav_depth'])) {
		// Init nav_depth
		$GLOBALS['nav_depth'] = '0';

		// Run the pre-filter chain
		$ret = runFilterChain('pre_youhere_line', array('access_level' => $accessLevel, 'type' => $type, 'search' => $search, 'prefix' => $prefix, 'link_add' => $linkAdd, 'content' => '', 'add' => $ADD));

		// Add pre-content
		$prefix = $ret['content'];

		// Add default content
		$prefix .= '<div class="you_are_here">{--YOU_ARE_HERE--}&nbsp;<strong><a class="you_are_here" href="{%url=modules.php?module=' . getModule() . $linkAdd . '%}">Home</a></strong>';
	} elseif ($return === FALSE) {
		// Count depth
		$GLOBALS['nav_depth']++;
	}

	// Add arrow
	$prefix .= '&nbsp;-&gt;&nbsp;';

	// We need to remove .php and the end
	if (substr($search, -4, 4) == '.php') {
		// Remove the .php
		$search = substr($search, 0, -4);
	} // END - if

	// Is ext-sql_patches installed?
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'isExtensionInstalledAndNewer()=' . intval(isExtensionInstalledAndNewer('sql_patches', '0.2.3')) . ',youre_here=' . getYoureHere() . ',isAdmin()=' . intval(isAdmin()) . ',modCheck=' . $modCheck);
	if (((isExtensionInstalledAndNewer('sql_patches', '0.2.3')) && (isYoureHereEnabled())) || ((isAdmin()) && ($modCheck == 'admin'))) {
		// Output HTML code
		$OUT = $prefix . '<strong><a class="you_are_here" href="{%url=modules.php?module=' . $modCheck . '&amp;' . $type . '=' . $search . $linkAdd . '%}">' . getTitleFromMenu($accessLevel, $search, $type, $ADD) . '</a></strong>';

		// Can we close the you-are-here navigation?
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'type=' . $type . ',getWhat()=' . getWhat() . ',accessLevel=' . $accessLevel . ',isWhatSet()=' . intval(isWhatSet()));
		if (($type == 'what') || (($type == 'action') && ((!isWhatSet()) || (($accessLevel == 'admin') && (getWhat() == 'welcome'))))) {
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'type=' . $type);
			// Add closing div and br-tag
			$GLOBALS['nav_depth'] = '0';

			// Run the post-filter chain
			$ret = runFilterChain('post_youhere_line', array('access_level' => $accessLevel, 'type' => $type, 'search' => $search, 'prefix' => $prefix, 'link_add' => $linkAdd, 'content' => $OUT, 'add' => $ADD));

			// Get content from filter back
			$OUT = $ret['content'];

			// Close div-tag, so not the filters have to do it
			$OUT .= '</div>';
		} // END - if
	} // END - if

	// Return or output HTML code?
	if ($return === TRUE) {
		// Return HTML code
		return $OUT;
	} else {
		// Output HTML code here
		outputHtml($OUT);
	}
}

// Adds a menu (mode = guest/member/admin/sponsor) to output
function addMenu ($mode, $action, $what) {
	// Init some variables
	$main_cnt = '0';

	// is the menu action valid?
	if (!isMenuActionValid($mode, $action, $what, TRUE)) {
		return getCode('MENU_NOT_VALID');
	} // END - if

	// Non-admin shall not see all menus
	$ADD = " AND `visible`='Y' AND `locked`='N'";
	if (isAdmin()) {
		// Is admin, so make all visible
		$ADD = '';
	} // END - if

	// Load SQL data and add the menu to the output stream...
	$result_main = sqlQueryEscaped("SELECT
	`title`,
	`what`,
	`action`,
	`visible`,
	`locked`
FROM
	`{?_MYSQL_PREFIX?}_%s_menu`
WHERE
	(`what`='' OR `what` IS NULL)
	".$ADD."
ORDER BY
	`sort` ASC",
		array($mode), __FUNCTION__, __LINE__);

	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'main_cnt=' . $main_cnt . ',getWhat()=' . getWhat());
	if (!ifSqlHasZeroNums($result_main)) {
		// There are menus available, so we simply display them... :)
		$GLOBALS['rows'] = '';
		while ($content = sqlFetchArray($result_main)) {
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'main_cnt=' . $main_cnt . ',action=' . $content['action'] . ',getWhat()=' . getWhat());
			// Disable the block-mode
			enableBlockMode(FALSE);

			// Load menu header template
			$GLOBALS['rows'] .= loadTemplate($mode . '_menu_title', TRUE, $content);

			// Sub menu
			$result_sub = sqlQueryEscaped("SELECT
	`title` AS `sub_title`,
	`what` AS `sub_what`,
	`visible` AS `sub_visible`,
	`locked` AS `sub_locked`
FROM
	`{?_MYSQL_PREFIX?}_%s_menu`
WHERE
	`action`='%s' AND
	`what` != '' AND
	`what` IS NOT NULL
	" . $ADD . "
ORDER BY
	`sort` ASC",
				array(
					$mode,
					$content['action']
				), __FUNCTION__, __LINE__);

			// Are there some entries?
			if (!ifSqlHasZeroNums($result_sub)) {
				// Init counter
				$count = '0';

				// Load all sub menus
				while ($content2 = sqlFetchArray($result_sub)) {
					// Merge both arrays in one
					$content = merge_array($content, $content2);

					// Init content
					$OUT = '';

					// Full file name for checking menu
					//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'sub_what=' . $content['sub_what']);
					$inc = sprintf('inc/modules/%s/what-%s.php', $mode, $content['sub_what']);
					if (isIncludeReadable($inc)) {
						// Mark currently selected menu - open
						if ((!empty($what)) && (($what == $content['sub_what']))) {
							$OUT = '<strong>';
						} // END - if

						// Is ext-sql_patches up-to-date, and display_home_in_index is Y?
						if ((getModule() == 'index') && (isExtensionInstalledAndNewer('sql_patches', '0.8.3')) && (isDisplayHomeInIndexEnabled()) && ($content['sub_what'] == getIndexHome())) {
							// Use index.php as link
							$OUT .= '<a name="menu" class="menu_blur" href="{%url=index.php%}" target="_self">';
						} else {
							// Regular navigation link
							$OUT .= '<a name="menu" class="menu_blur" href="{%url=modules.php?module=' . getModule() . '&amp;what=' . $content['sub_what'] . '%}" target="_self">';
						}
					} else {
						// Not found - open
						$OUT .= '<span class="bad" style="cursor:help" title="{%message,ADMIN_MENU_WHAT_404_TITLE=' . $content['sub_what'] . '%}">';
					}

					// Menu title
					$OUT .= '{?menu_blur_spacer?}' . $content['sub_title'];

					if (isIncludeReadable($inc)) {
						$OUT .= '</a>';

						// Mark currently selected menu - close
						if ((!empty($what)) && (($what == $content['sub_what']))) {
							$OUT .= '</strong>';
						} // END - if
					} else {
						// Not found - close
						$OUT .= '</span>';
					}

					// Cunt it up
					$count++;

					// Rewrite array
					$content = array(
						'menu'    => $OUT,
						'what'    => $content['sub_what'],
						'visible' => $content['sub_visible'],
						'locked'  => $content['locked'],
					);

					// Add regular menu row or bottom row?
					if ($count < sqlNumRows($result_sub)) {
						$GLOBALS['rows'] .= loadTemplate($mode . '_menu_row', TRUE, $content);
					} else {
						$GLOBALS['rows'] .= loadTemplate($mode . '_menu_bottom', TRUE, $content);
					}
				} // END - while
			} else {
				// This is a menu block... ;-)
				enableBlockMode();

				// Load menu block
				$INC = sprintf('inc/modules/%s/action-%s.php', $mode, $content['action']);
				if (isFileReadable($INC)) {
					// Load include file
					if ((!isExtensionActive($content['action'])) || ($content['action'] == 'online')) $GLOBALS['rows'] .= loadTemplate('menu_what_begin', TRUE, $mode);
					//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'main_cnt=' . $main_cnt . ',action=' . $content['action'] . ',getWhat()=' . getWhat());
					loadInclude($INC);
					//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'main_cnt=' . $main_cnt . ',action=' . $content['action'] . ',getWhat()=' . getWhat());
					if ((!isExtensionActive($content['action'])) || ($content['action'] == 'online')) $GLOBALS['rows'] .= loadTemplate('menu_what_end', TRUE, $mode);
				}
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'main_cnt=' . $main_cnt . ',action=' . $content['action'] . ',getWhat()=' . getWhat());
			}

			// Free result
			sqlFreeResult($result_sub);

			// Count one up
			$main_cnt++;

			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'main_cnt=' . $main_cnt . ',getWhat()=' . getWhat());
			if (sqlNumRows($result_main) > $main_cnt) {
				// Add separator
				$GLOBALS['rows'] .= loadTemplate('menu_separator', TRUE, $mode);

				// Prepare filter data array
				$filterData = array(
					'output'    => '',
					'exclusive' => FALSE,
					'action'    => $action,
					'mode'      => $mode,
				);

				// Run filter chain
				$filterData = runFilterChain($mode . '_menu_advert', $filterData);

				// Add content output
				$GLOBALS['rows'] .= $filterData['output'];
			} // END - if
		} // END - while

		// Free memory
		sqlFreeResult($result_main);

		// Prepare filter data array
		$filterData = array(
			'output'    => '',
			'exclusive' => FALSE,
			'action'    => $action,
			'mode'      => $mode,
		);

		// Run filter chain
		$filterData = runFilterChain($mode . '_menu_advert_end', $filterData);

		// Add content output
		$GLOBALS['rows'] .= $filterData['output'];

		// Prepare data
		$content = array(
			'rows'      => $GLOBALS['rows'],
			'menu_mode' => $mode
		);

		// Load main template
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'main_cnt=' . $main_cnt . ',getWhat()=' . getWhat());
		loadTemplate('menu_table', FALSE, $content);
	} // END - if
}

// Checks whether the current user is a member
function isMember () {
	// By default no member
	$ret = FALSE;

	// Fix missing 'last_online' array, damn stupid code :(((
	// @TODO Try to rewrite this to one or more functions
	if ((!isset($GLOBALS['last_online'])) || (!is_array($GLOBALS['last_online']))) {
		$GLOBALS['last_online'] = array();
	} // END - if

	// Is the cache entry there?
	if (isset($GLOBALS[__FUNCTION__])) {
		// Then return it
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'CACHED! (' . intval($GLOBALS[__FUNCTION__]) . ')');
		return $GLOBALS[__FUNCTION__];
	} elseif ((!isSessionVariableSet('userid')) || (!isSessionVariableSet('u_hash'))) {
		// Destroy any existing user session data
		destroyMemberSession();
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'No member set in cookie/session.');

		// Abort further processing
		return FALSE;
	}

	// Get userid secured from session
	setMemberId(getSession('userid'));

	// ... and set it as currently handled user id
	setCurrentUserId(getMemberId());

	// Init user data array
	initUserData();

	// Fix "deleted" cookies
	fixDeletedCookies(array('userid', 'u_hash'));

	// Are cookies set and can the member data be loaded?
	if ((isMemberIdSet()) && (isSessionVariableSet('u_hash')) && (fetchUserData(getMemberId()) === TRUE)) {
		// Validate password by created the difference of it and the secret key
		$valPass = encodeHashForCookie(getUserData('password'));

		// So did we now have valid data and an unlocked user?
		if ((getUserData('status') == 'CONFIRMED') && ($valPass == getSession('u_hash'))) {
			// Transfer last module and online time
			$GLOBALS['last_online']['module'] = getUserData(getUserLastWhatName());
			$GLOBALS['last_online']['online'] = getUserData('last_online');

			// Account is confirmed and all cookie data is valid so he is definely logged in! :-)
			$ret = TRUE;
		} // END - if
	} // END - if

	// Is $ret still false?
	if ($ret === FALSE) {
		// Yes, so destroy the session
		destroyMemberSession();
	} // END - if

	// Cache status
	$GLOBALS[__FUNCTION__] = $ret;

	// Return status
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'ret=' . intval($ret));
	return $ret;
}

// Fetch user data for given user id
function fetchUserData ($value, $column = 'userid') {
	// Extension ext-user must be there at any case
	if (!isExtensionActive('user')) {
		// Absent ext-user is really not good
		return FALSE;
	} elseif (is_null($value)) {
		// This shall never happen, so please report it
		reportBug(__FUNCTION__, __LINE__, 'value=NULL,column=' . $column . ' - value can never be NULL');
	}

	// If we should look for userid secure&set it here
	if (substr($column, -2, 2) == 'id') {
		// Secure userid
		$value = bigintval($value);

		// Don't look for invalid userids...
		if (!isValidId($value)) {
			// Invalid, so abort here
			reportBug(__FUNCTION__, __LINE__, 'User id ' . $value . ' is invalid.');
		} // END - if

		// Unset cached values if found and different
		if ((isCurrentUserIdSet()) && (getCurrentUserId() != $value)) {
			// Unset it
			unsetCurrentUserId();
		} elseif (isValidUserData()) {
			// Use cache, so it is fine
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'value=' . $value . ' is valid, using cache #1');
			return TRUE;
		} // END - if
	} elseif (isValidUserData()) {
		// Using cache is fine
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'value=' . $value . ' is valid, using cache #2');
		return TRUE;
	}

	// By default none was found
	$found = FALSE;

	// Extra SQL statements
	$ADD = runFilterChain('convert_user_data_columns', ' ');

	// Query for the user
	$result = sqlQueryEscaped("SELECT *" . $ADD . " FROM `{?_MYSQL_PREFIX?}_user_data` WHERE `%s`='%s' LIMIT 1",
		array(
			$column,
			$value
		), __FUNCTION__, __LINE__);

	// Is there a record?
	if (sqlNumRows($result) == 1) {
		// Load data from cookies
		$data = sqlFetchArray($result);

		// Set the userid for later use
		setCurrentUserId($data['userid']);

		// And cache the data for this userid
		$GLOBALS['user_data'][getCurrentUserId()] = $data;

		// Rewrite 'last_failure' if found and ext-user has version >= 0.3.7
		if ((isExtensionInstalledAndNewer('user', '0.3.7')) && (isset($GLOBALS['user_data'][getCurrentUserId()]['last_failure']))) {
			// Backup the raw one and zero it
			$GLOBALS['user_data'][getCurrentUserId()]['last_failure_raw'] = $GLOBALS['user_data'][getCurrentUserId()]['last_failure'];
			$GLOBALS['user_data'][getCurrentUserId()]['last_failure'] = NULL;

			// Is it not zero?
			if (!is_null($GLOBALS['user_data'][getCurrentUserId()]['last_failure_raw'])) {
				// Seperate data/time
				$array = explode(' ', $GLOBALS['user_data'][getCurrentUserId()]['last_failure_raw']);

				// Seperate data and time again
				$array['date'] = explode('-', $array[0]);
				$array['time'] = explode(':', $array[1]);

				// Now pass it to mktime()
				$GLOBALS['user_data'][getCurrentUserId()]['last_failure'] = mktime(
					$array['time'][0],
					$array['time'][1],
					$array['time'][2],
					$array['date'][1],
					$array['date'][2],
					$array['date'][0]
				);
			} // END - if
		} // END - if

		// Found, but valid?
		$found = isValidUserData();
	} // END - if

	// Free memory
	sqlFreeResult($result);

	// Return result
	return $found;
}

/*
 * Checks whether the current session bears a valid admin id and password hash.
 *
 * This patched function will reduce many SELECT queries for the current admin
 * login.
 */
function isAdmin () {
	// Is there cache?
	if (isset($GLOBALS[__FUNCTION__])) {
		// Return it
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'isAdmin()=' . intval($GLOBALS[__FUNCTION__]));
		return $GLOBALS[__FUNCTION__];
	} // END - if

	// No admin in installation phase!
	if ((isInstaller()) || (!isAdminRegistered())) {
		$GLOBALS[__FUNCTION__] = FALSE;
		return FALSE;
	} // END - if

	// Init variables
	$ret = FALSE;
	$adminId = '0';
	$passwordFromCookie = '';
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, $adminId);

	// If admin login is not given take current from cookies...
	if ((isSessionVariableSet('admin_id')) && (isSessionVariableSet('admin_md5'))) {
		// Get admin login and password from session/cookies
		$adminId            = getCurrentAdminId();
		$passwordFromCookie = getAdminMd5();
	} // END - if
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'adminId=' . $adminId . ',passwordFromCookie=' . $passwordFromCookie);

	// Abort if admin id is zero
	if (($adminId == '0') || (empty($passwordFromCookie))) {
		// A very noisy debug message ...
		//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Current adminId is zero. isSessionVariableSet(admin_id)=' . intval(isSessionVariableSet('admin_id')) . ',isSessionVariableSet(admin_md5)=' . intval(isSessionVariableSet('admin_md5')));

		// Abort here now
		$GLOBALS[__FUNCTION__] = FALSE;
		return FALSE;
	} // END - if

	// Init it with failed
	$GLOBALS[__FUNCTION__] = FALSE;

	// Search in array for entry
	if (isset($GLOBALS['admin_hash'])) {
		// Use cached string
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Using admin_hash=' . $GLOBALS['admin_hash'] . ' from cache');
	} elseif ((!empty($adminId)) && (!empty($passwordFromCookie)) && (isAdminHashSet($adminId) === TRUE)) {
		// Get admin hash and hash it
		$GLOBALS['admin_hash'] = encodeHashForCookie(getAdminHash($adminId));
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'valPass=' . $GLOBALS['admin_hash']);

		// Count cache hits
		incrementStatsEntry('cache_hits');
	} elseif ((!empty($adminId)) && ((!isExtensionActive('cache')) || (isAdminHashSet($adminId) === FALSE))) {
		// Get admin hash and hash it
		$GLOBALS['admin_hash'] = encodeHashForCookie(getAdminHash($adminId));
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'valPass=' . $GLOBALS['admin_hash']);
	}

	// Check if password is valid
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, '(' . $GLOBALS['admin_hash'] . '==' . $passwordFromCookie . ')='.intval($GLOBALS['admin_hash'] == $passwordFromCookie));
	$GLOBALS[__FUNCTION__] = ((!empty($GLOBALS['admin_hash'])) && ($GLOBALS['admin_hash'] == $passwordFromCookie));

	// Return result of comparision
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'isAdmin()=' . intval($GLOBALS[__FUNCTION__]));
	return $GLOBALS[__FUNCTION__];
}

// Generates a list of "max receiveable emails per day"
function addMaxReceiveList ($mode, $default = '') {
	$OUT = '';
	$result = FALSE;

	switch ($mode) {
		case 'guest':
			// Guests (in the registration form) are not allowed to select 0 mails per day.
			$result = sqlQuery('SELECT `value`, `comment` FROM `{?_MYSQL_PREFIX?}_max_receive` WHERE `value` > 0 ORDER BY `value` ASC',
			__FUNCTION__, __LINE__);
			break;

		case 'admin':
		case 'member':
			// Members are allowed to set to zero mails per day (we will change this soon!)
			$result = sqlQuery('SELECT `value`, `comment` FROM `{?_MYSQL_PREFIX?}_max_receive` ORDER BY `value` ASC',
			__FUNCTION__, __LINE__);
			break;

		default: // Invalid!
			logDebugMessage(__FUNCTION__, __LINE__, sprintf('Invalid mode %s detected.', $mode));
			break;
	}

	// Some entries are found?
	if (!ifSqlHasZeroNums($result)) {
		$OUT = '';
		while ($content = sqlFetchArray($result)) {
			$OUT .= '      <option value="' . $content['value'] . '"';

			if (postRequestElement('max_mails') == $content['value']) {
				$OUT .= ' selected="selected"';
			} // END - if

			$OUT .= '>' . $content['value'] . ' {--PER_DAY--}';
			if (!empty($content['comment'])) $OUT .= '(' . $content['comment'] . ')';
			$OUT .= '</option>';
		}

		// Load template
		$OUT = loadTemplate(($mode . '_receive_table'), TRUE, $OUT);
	} else {
		// Maybe the admin has to setup some maximum values?
		reportBug(__FUNCTION__, __LINE__, 'Nothing is being done here?');
	}

	// Free result
	sqlFreeResult($result);

	// Return generated HTML code
	return $OUT;
}

// Checks whether the given email address is used.
function isEmailTaken ($email) {
	// Default is no userid
	$useridSql = ' IS NOT NULL';

	// Is a member logged in?
	if (isMember()) {
		// Get userid
		$useridSql = '!= ' . bigintval(getMemberId());
	} // END - if

	// Replace dot with {DOT}
	$email = str_replace('.', '{DOT}', $email);

	// Query the database
	$result = sqlQueryEscaped("SELECT
	COUNT(`userid`) AS `cnt`
FROM
	`{?_MYSQL_PREFIX?}_user_data`
WHERE
	'%s' REGEXP `email` AND
	`userid` %s
LIMIT 1",
		array(
			$email,
			$useridSql
		), __FUNCTION__, __LINE__);

	// Is the email there?
	list($count) = sqlFetchRow($result);

	// Free the result
	sqlFreeResult($result);

	// Return result
	return ($count == 1);
}

// Validate the given menu action
function isMenuActionValid ($mode, $action, $what, $updateEntry = FALSE) {
	// Is the cache entry there and we shall not update?
	if ((isset($GLOBALS['action_valid'][$mode][$action][$what])) && ($updateEntry === FALSE)) {
		// Count cache hit
		incrementStatsEntry('cache_hits');

		// Then use this cache
		return $GLOBALS['action_valid'][$mode][$action][$what];
	} // END - if

	// By default nothing is valid
	$ret = FALSE;

	// Look in all menus or only unlocked
	$add = '';
	if ((!isAdmin()) && ($mode != 'admin')) $add = " AND `locked`='N'";

	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'mode=' . $mode . ',action=' . $action . ',what=' . $what);
	if (($mode != 'admin') && ($updateEntry === TRUE)) {
		// Update guest or member menu
		$sql = sqlQueryEscaped("UPDATE `{?_MYSQL_PREFIX?}_%s_menu` SET `counter`=`counter`+1 WHERE `action`='%s' AND `what`='%s'".$add." LIMIT 1",
			array(
				$mode,
				$action,
				$what
			), __FUNCTION__, __LINE__, FALSE);
	} elseif (($what != 'welcome') && (!empty($what))) {
		// Other actions
		$sql = sqlQueryEscaped("SELECT `id`, `what` FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE `action`='%s' AND `what`='%s'".$add." ORDER BY `action` DESC LIMIT 1",
			array(
				$mode,
				$action,
				$what
			), __FUNCTION__, __LINE__, FALSE);
	} else {
		// Admin login overview
		$sql = sqlQueryEscaped("SELECT `id`, `what` FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE `action`='%s' AND (`what`='' OR `what` IS NULL)".$add." ORDER BY `action` DESC LIMIT 1",
			array(
				$mode,
				$action
			), __FUNCTION__, __LINE__, FALSE);
	}

	// Run SQL command
	$result = sqlQuery($sql, __FUNCTION__, __LINE__);

	// Should we look for affected rows (only update) or found rows?
	if ($updateEntry === TRUE) {
		// Check updated/affected rows
		$ret = (!ifSqlHasZeroAffectedRows());
	} else {
		// Check found rows
		$ret = (!ifSqlHasZeroNums($result));
	}

	// Free memory
	sqlFreeResult($result);

	// Set cache entry
	$GLOBALS['action_valid'][$mode][$action][$what] = $ret;

	// Return result
	return $ret;
}

// Get action value from mode (admin/guest/member) and what-value
function getActionFromModuleWhat ($module, $what) {
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'module=' . $module . ',what=' . $what);
	// Init status
	$data['action'] = '';

	if (!isExtensionInstalledAndNewer('sql_patches', '0.0.5')) {
		// ext-sql_patches is missing so choose depending on mode
		$what = determineWhat($module);
	} elseif ((empty($what)) && ($module != 'admin')) {
		// Use configured 'home'
		$what = getIndexHome();
	} // END - if

	if ($module == 'admin') {
		// Action value for admin area
		if (isGetRequestElementSet('action')) {
			// Use from request!
			return getRequestElement('action');
		} elseif (isActionSet()) {
			// Get it directly from URL
			return getAction();
		} elseif (($what == 'welcome') || (!isWhatSet())) {
			// Default value for admin area
			$data['action'] = 'login';
		}
	} elseif (isActionSet()) {
		// Get it directly from URL
		return getAction();
	}
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, ' ret=' . $data['action']);

	// Does the module have a menu?
	if (ifModuleHasMenu($module)) {
		// Rewriting modules to menu
		$module = mapModuleToTable($module);

		// Guest and member menu is 'main' as the default
		if (empty($data['action'])) {
			$data['action'] = 'main';
		} // END - if

		// Load from database
		$result = sqlQueryEscaped("SELECT `action` FROM `{?_MYSQL_PREFIX?}_%s_menu` WHERE `what`='%s' LIMIT 1",
			array(
				$module,
				$what
			), __FUNCTION__, __LINE__);

		// Entry found?
		if (sqlNumRows($result) == 1) {
			// Load action value and pray that this one is the right you want... ;-)
			$data = sqlFetchArray($result);
		} // END - if

		// Free memory
		sqlFreeResult($result);
	} elseif ((!isExtensionInstalled('sql_patches')) && ($module != 'admin') && ($module != 'unknown')) {
		// No ext-sql_patches installed, but maybe we need to register an admin?
		if (isAdminRegistered()) {
			// Redirect to admin area
			redirectToUrl('admin.php');
		} // END - if
	}

	// Return action value
	return $data['action'];
}

// Get category name back
function getCategory ($cid) {
	// Default is not found
	$data['cat'] = '{--_CATEGORY_404--}';

	// Is the category id set?
	if (!isValidId($cid)) {
		// No category
		$data['cat'] = '{--_CATEGORY_NONE--}';
	} elseif (isValidId($cid)) {
		// Lookup the category in database
		$result = sqlQueryEscaped('SELECT `cat` FROM `{?_MYSQL_PREFIX?}_cats` WHERE `id`=%s LIMIT 1',
			array(bigintval($cid)), __FUNCTION__, __LINE__);
		if (sqlNumRows($result) == 1) {
			// Category found... :-)
			$data = sqlFetchArray($result);
		} // END - if

		// Free result
		sqlFreeResult($result);
	} // END - if

	// Return result
	return $data['cat'];
}

// Get a string of "mail title" and price back
function getPaymentTitlePrice ($paymentsId, $full = FALSE) {
	// Only title or also including price?
	if ($full === FALSE) {
		$ret = getPaymentData($paymentsId, 'main_title');
	} else {
		$ret = getPaymentData($paymentsId, 'main_title') . ' / {%pipe,getPaymentData,translateComma=' . $paymentsId . '%} {?POINTS?}';
	}

	// Return result
	return $ret;
}

// Get payment price
function getPaymentPrice ($paymentsId) {
	// Return result
	return getPaymentData($paymentsId, 'price');
}

// Get payment time
function getPaymentTime ($paymentsId) {
	// Return result
	return getPaymentData($paymentsId, 'time');
}

// Get payment 'payment'
function getPaymentPayment ($paymentsId) {
	// Return result
	return getPaymentData($paymentsId, 'payment');
}

// "Getter" for payment data (cached)
function getPaymentData ($paymentsId, $lookFor) {
	// Default value...
	$data[$lookFor] = NULL;

	// Is there cache?
	if (isset($GLOBALS['cache_array']['payments'][$lookFor][$paymentsId])) {
		// Use it if found to save SQL queries
		$data[$lookFor] = $GLOBALS['cache_array']['payments'][$lookFor][$paymentsId];

		// Update cache hits
		incrementStatsEntry('cache_hits');
	} elseif (!isExtensionActive('cache')) {
		// Search for it in database
		$result = sqlQueryEscaped('SELECT `%s` FROM `{?_MYSQL_PREFIX?}_payments` WHERE `id`=%s LIMIT 1',
			array(
				$lookFor,
				bigintval($paymentsId)
			), __FUNCTION__, __LINE__);

		// Is the entry there?
		if (sqlNumRows($result) == 1) {
			// Payment type found... :-)
			$data = sqlFetchArray($result);
		} // END - if

		// Free result
		sqlFreeResult($result);
	}

	// Return value
	return $data[$lookFor];
}

// Remove a receiver's id from $receivers and add a link for him to confirm
function removeReceiver (&$receivers, $key, $userid, $poolId, $statsId = 0, $isBonusMail = FALSE) {
	// Default is not removed
	$ret = 'failed';

	// Is the userid valid?
	if (isValidId($userid)) {
		// Remove entry from array
		unset($receivers[$key]);

		// Is there already a line for this user available?
		if (isValidId($statsId)) {
			// Default is 'normal' mail
			$type = 'NORMAL';
			$rowName = 'stats_id';

			// Only when we got a real stats id continue searching for the entry
			if ($isBonusMail === TRUE) {
				$type = 'BONUS';
				$rowName = 'bonus_id';
			} // END - if

			// Try to look the entry up
			$result = sqlQueryEscaped("SELECT `id` FROM `{?_MYSQL_PREFIX?}_user_links` WHERE `%s`=%s AND `userid`=%s AND `link_type`='%s' LIMIT 1",
				array(
					$rowName,
					bigintval($statsId),
					bigintval($userid),
					$type
				), __FUNCTION__, __LINE__);

			// Was it *not* found?
			if (ifSqlHasZeroNums($result)) {
				// So we add one!
				sqlQueryEscaped("INSERT INTO `{?_MYSQL_PREFIX?}_user_links` (`%s`, `userid`, `link_type`) VALUES (%s,%s,'%s')",
					array(
						$rowName,
						bigintval($statsId),
						bigintval($userid),
						$type
					), __FUNCTION__, __LINE__);

				// Update 'mails_sent' if ext-sql_patches is updated
				if (isExtensionInstalledAndNewer('sql_patches', '0.7.4')) {
					// Update the pool
					updatePoolDataById($poolId, 'mails_sent', 1, '+');
				} // END - if
				$ret = 'done';
			} else {
				// Already found
				$ret = 'already';
			}

			// Free memory
			sqlFreeResult($result);
		} // END - if
	} // END - if

	// Return status for sending routine
	return $ret;
}

// Calculate sum (default) or count records of given criteria
function countSumTotalData ($search, $tableName, $lookFor = 'id', $whereStatement = 'userid', $countRows = FALSE, $add = '', $mode = '=') {
	// Debug message
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'search=' . $search . ',tableName=' . $tableName . ',lookFor=' . $lookFor . ',whereStatement=' . $whereStatement . ',add=' . $add);
	if ((empty($search)) && (!is_null($search))) {
		// Count or sum whole table?
		if ($countRows === TRUE) {
			// Count whole table
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'COUNT!');
			$result = sqlQueryEscaped('SELECT COUNT(`%s`) AS `res` FROM `{?_MYSQL_PREFIX?}_%s`' . $add . ' LIMIT 1',
				array(
					$lookFor,
					$tableName
				), __FUNCTION__, __LINE__);
		} else {
			// Sum whole table
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'SUM!');
			$result = sqlQueryEscaped('SELECT SUM(`%s`) AS `res` FROM `{?_MYSQL_PREFIX?}_%s`' . $add . ' LIMIT 1',
				array(
					$lookFor,
					$tableName
				), __FUNCTION__, __LINE__);
		}
	} elseif (($countRows === TRUE) || ($lookFor == 'userid')) {
		// Count rows
		if (is_null($search)) {
			// Fix mode
			if ($mode == '=') {
				$mode = 'IS';
			} elseif ($mode == '!=') {
				$mode = 'IS NOT';
			}

			// Look for NULL/0
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'COUNT-NULL!');
			$result = sqlQueryEscaped("SELECT COUNT(`%s`) AS `res` FROM `{?_MYSQL_PREFIX?}_%s` WHERE (`%s` %s NULL OR `%s`=0)" . $add . ' LIMIT 1',
				array(
					$lookFor,
					$tableName,
					$whereStatement,
					$mode,
					$whereStatement
				), __FUNCTION__, __LINE__);
		} else {
			// Regular entry
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'COUNT!');
			$result = sqlQueryEscaped("SELECT COUNT(`%s`) AS `res` FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s`%s'%s'" . $add . ' LIMIT 1',
				array(
					$lookFor,
					$tableName,
					$whereStatement,
					$mode,
					$search
				), __FUNCTION__, __LINE__);
		}
	} else {
		// Sum all rows
		if (is_null($search)) {
			// Fix mode
			if ($mode == '=') {
				$mode = 'IS';
			} elseif ($mode == '!=') {
				$mode = 'IS NOT';
			}

			// Look for NULL/0
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'SUM-NULL!');
			$result = sqlQueryEscaped("SELECT SUM(`%s`) AS `res` FROM `{?_MYSQL_PREFIX?}_%s` WHERE (`%s` %s NULL OR `%s`=0)" . $add . ' LIMIT 1',
				array(
					$lookFor,
					$tableName,
					$whereStatement,
					$mode,
					$whereStatement
				), __FUNCTION__, __LINE__);
		} else {
			// Regular entry
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'SUM!');
			$result = sqlQueryEscaped("SELECT SUM(`%s`) AS `res` FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s`%s'%s'" . $add . ' LIMIT 1',
				array(
					$lookFor,
					$tableName,
					$whereStatement,
					$mode,
					$search
				), __FUNCTION__, __LINE__);
		}
	}

	// Load row
	$data = sqlFetchArray($result);

	// Free result
	sqlFreeResult($result);

	// Fix empty values
	if ((empty($data['res'])) && ($lookFor != 'counter') && ($lookFor != 'id') && ($lookFor != 'userid') && ($lookFor != 'rallye_id')) {
		// Float number
		$data['res'] = '0.00000';
	} elseif ('' . $data['res'] . '' == '') {
		// Fix empty result
		$data['res'] = '0';
	}

	// Return value
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'res=' . $data['res']);
	return $data['res'];
}

/**
 * Sends out mail to all administrators. This function is no longer obsolete
 * because we need it when there is no ext-admins installed
 */
function sendAdminEmails ($subject, $message, $isBugReport = FALSE) {
	// Default is no special header
	$mailHeader = '';

	// Is it a bug report?
	if ($isBugReport === TRUE) {
		// Then add a reply-to line back to the author (me)
		$mailHeader = 'Reply-To: webmaster@mxchange.org' . PHP_EOL;
	} // END - if

	// Load all admin email addresses
	$result = sqlQuery('SELECT `email` FROM `{?_MYSQL_PREFIX?}_admins` ORDER BY `id` ASC', __FUNCTION__, __LINE__);

	// And send the notification to all of them
	while ($content = sqlFetchArray($result)) {
		// Send the email out
		sendEmail($content['email'], $subject, $message, 'N', $mailHeader);
	} // END - if

	// Free result
	sqlFreeResult($result);

	// Really simple... ;-)
}

// Get id number from administrator's login name
function getAdminId ($adminLogin) {
	// By default no admin is found
	$data['id'] = -1;

	// Check cache
	if (isset($GLOBALS['cache_array']['admin']['admin_id'][$adminLogin])) {
		// Use it if found to save SQL queries
		$data['id'] = $GLOBALS['cache_array']['admin']['admin_id'][$adminLogin];

		// Update cache hits
		incrementStatsEntry('cache_hits');
	} elseif (!isExtensionActive('cache')) {
		// Load from database
		$result = sqlQueryEscaped("SELECT `id` FROM `{?_MYSQL_PREFIX?}_admins` WHERE `login`='%s' LIMIT 1",
			array($adminLogin), __FUNCTION__, __LINE__);

		// Is there an entry?
		if (sqlNumRows($result) == 1) {
			// Get it
			$data = sqlFetchArray($result);
		} // END - if

		// Free result
		sqlFreeResult($result);
	}

	// Return the id
	return $data['id'];
}

// "Getter" for current admin id
function getCurrentAdminId () {
	// Log debug message
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'called!');

	// Is there cache?
	if (!isset($GLOBALS['current_admin_id'])) {
		// Get the admin login from session
		$adminId = getSession('admin_id');

		// Remember in cache securely
		setCurrentAdminId(bigintval($adminId));
	} // END - if

	// Return it
	return $GLOBALS['current_admin_id'];
}

// Setter for current admin id
function setCurrentAdminId ($currentAdminId) {
	// Set it secured
	$GLOBALS['current_admin_id'] = bigintval($currentAdminId);
}

// Get password hash from administrator's login name
function getAdminHash ($adminId) {
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'adminId=' . $adminId . ' - ENTERED!');
	// By default an invalid hash is returned
	$data['password'] = -1;

	// Is admin hash set?
	if (isAdminHashSet($adminId)) {
		// Check cache
		$data['password'] = $GLOBALS['cache_array']['admin']['password'][$adminId];

		// Update cache hits
		incrementStatsEntry('cache_hits');
	} elseif (!isExtensionActive('cache')) {
		// Load from database
		$result = sqlQueryEscaped("SELECT `password` FROM `{?_MYSQL_PREFIX?}_admins` WHERE `id`=%s LIMIT 1",
			array(bigintval($adminId)), __FUNCTION__, __LINE__);

		// Is there an entry?
		if (sqlNumRows($result) == 1) {
			// Fetch data
			$data = sqlFetchArray($result);

			// Set cache
			setAdminHash($adminId, $data['password']);
		} // END - if

		// Free result
		sqlFreeResult($result);
	}

	// Return password hash
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'adminId=' . $adminId . ',data[password]=' . $data['password'] . ' - EXIT!');
	return $data['password'];
}

// "Getter" for admin login
function getAdminLogin ($adminId) {
	// By default a non-existent login is returned (other functions react on this!)
	$data['login'] = '***';

	if (isset($GLOBALS['cache_array']['admin']['login'][$adminId])) {
		// Get cache
		$data['login'] = $GLOBALS['cache_array']['admin']['login'][$adminId];

		// Update cache hits
		incrementStatsEntry('cache_hits');
	} elseif (!isExtensionActive('cache')) {
		// Load from database
		$result = sqlQueryEscaped("SELECT `login` FROM `{?_MYSQL_PREFIX?}_admins` WHERE `id`=%s LIMIT 1",
			array(bigintval($adminId)), __FUNCTION__, __LINE__);

		// Entry found?
		if (sqlNumRows($result) == 1) {
			// Fetch data
			$data = sqlFetchArray($result);

			// Set cache
			$GLOBALS['cache_array']['admin']['login'][$adminId] = $data['login'];
		} // END - if

		// Free memory
		sqlFreeResult($result);
	}

	// Return the result
	return $data['login'];
}

// Get email address of admin id
function getAdminEmail ($adminId) {
	// By default an invalid emails is returned
	$data['email'] = '***';

	if (isset($GLOBALS['cache_array']['admin']['email'][$adminId])) {
		// Get cache
		$data['email'] = $GLOBALS['cache_array']['admin']['email'][$adminId];

		// Update cache hits
		incrementStatsEntry('cache_hits');
	} elseif (!isExtensionActive('cache')) {
		// Load from database
		$result_admin_id = sqlQueryEscaped("SELECT `email` FROM `{?_MYSQL_PREFIX?}_admins` WHERE `id`=%s LIMIT 1",
			array(bigintval($adminId)), __FUNCTION__, __LINE__);

		// Entry found?
		if (sqlNumRows($result_admin_id) == 1) {
			// Get data
			$data = sqlFetchArray($result_admin_id);

			// Set cache
			$GLOBALS['cache_array']['admin']['email'][$adminId] = $data['email'];
		} // END - if

		// Free result
		sqlFreeResult($result_admin_id);
	}

	// Return email
	return $data['email'];
}

// Get default ACL of admin id
function getAdminDefaultAcl ($adminId) {
	// By default an invalid ACL value is returned
	$data['default_acl'] = 'NO-ACL';

	// Is ext-sql_patches there and was it found in cache?
	if (!isExtensionActive('sql_patches')) {
		// Not found, which is bad, so we need to allow all
		$data['default_acl'] = 'allow';
	} elseif (isset($GLOBALS['cache_array']['admin']['default_acl'][$adminId])) {
		// Use cache
		$data['default_acl'] = $GLOBALS['cache_array']['admin']['default_acl'][$adminId];

		// Update cache hits
		incrementStatsEntry('cache_hits');
	} elseif (!isExtensionActive('cache')) {
		// Load from database
		$result_admin_id = sqlQueryEscaped("SELECT `default_acl` FROM `{?_MYSQL_PREFIX?}_admins` WHERE `id`=%s LIMIT 1",
			array(bigintval($adminId)), __FUNCTION__, __LINE__);

		// Is there an entry?
		if (sqlNumRows($result_admin_id) == 1) {
			// Fetch data
			$data = sqlFetchArray($result_admin_id);

			// Set cache
			$GLOBALS['cache_array']['admin']['default_acl'][$adminId] = $data['default_acl'];
		}

		// Free result
		sqlFreeResult($result_admin_id);
	}

	// Return default ACL
	return $data['default_acl'];
}

// Get menu mode (la_mode) of admin id
function getAdminMenuMode ($adminId) {
	// By default an invalid mode
	$data['la_mode'] = 'INVALID';

	// Is ext-sql_patches there and was it found in cache?
	if (!isExtensionActive('sql_patches')) {
		// Not found, which is bad, so we need to allow all
		$data['la_mode'] = 'global';
	} elseif (isset($GLOBALS['cache_array']['admin']['la_mode'][$adminId])) {
		// Use cache
		$data['la_mode'] = $GLOBALS['cache_array']['admin']['la_mode'][$adminId];

		// Update cache hits
		incrementStatsEntry('cache_hits');
	} elseif (!isExtensionActive('cache')) {
		// Load from database
		$result_admin_id = sqlQueryEscaped("SELECT `la_mode` FROM `{?_MYSQL_PREFIX?}_admins` WHERE `id`=%s LIMIT 1",
			array(bigintval($adminId)), __FUNCTION__, __LINE__);

		// Is there an entry?
		if (sqlNumRows($result_admin_id) == 1) {
			// Fetch data
			$data = sqlFetchArray($result_admin_id);

			// Set cache
			$GLOBALS['cache_array']['admin']['la_mode'][$adminId] = $data['la_mode'];
		}

		// Free result
		sqlFreeResult($result_admin_id);
	}

	// Return default ACL
	return $data['la_mode'];
}

// Generates an option list from various parameters
function generateOptions ($table, $key, $value, $default = '', $extra = '', $whereStatement = '', $disabled = array(), $callback = '', $allowNone = FALSE) {
	$ret = '';

	// Allow none?
	if ($allowNone === TRUE) {
		// Add option for none
		$ret .= '<option value="0">{--SELECT_NONE--}</option>';
	} // END - if

	if ($table == '/ARRAY/') {
		// Selection from array
		if ((is_array($key)) && (is_array($value)) && ((count($key)) == (count($value)) || (!empty($callback)))) {
			// Both are arrays
			foreach ($key as $idx => $optionValue) {
				$ret .= '<option value="' . $optionValue . '"';
				if ($default == $optionValue) {
					// Selected by default
					$ret .= ' selected="selected"';
				} elseif (isset($disabled[$optionValue])) {
					// Disabled!
					$ret .= ' disabled="disabled"';
				}

				// Is the call-back function set?
				if (!empty($callback)) {
					// Call it
					$value[$idx] = call_user_func_array($callback, array($key[$idx]));
				} // END - if

				// Finish option tag
				$ret .= '>' . $value[$idx] . '</option>';
			} // END - foreach
		} else {
			// Problem in request
			reportBug(__FUNCTION__, __LINE__, 'Not all are arrays: key[' . count($key) . ']=' . gettype($key) . ',value[' . count($value) . ']=' . gettype($value) . ',callback=' . $callback);
		}
	} else {
	    ///////////////////////
		// Data from database /
		///////////////////////

		// Init extra column (if requested)
		$extraColumn = '';
		if (!empty($extra)) {
			$extraColumn = ',`' . $extra . '` AS `extra`';
		} // END - if

		// Run SQL query
		$result = sqlQueryEscaped("SELECT `%s` AS `key`, `%s` AS `value`" . $extraColumn . " FROM `{?_MYSQL_PREFIX?}_%s` " . $whereStatement . " ORDER BY `%s` ASC",
			array(
				$key,
				$value,
				$table,
				$value
			), __FUNCTION__, __LINE__);

		// Is there rows?
		if (!ifSqlHasZeroNums($result)) {
			// Found data so add them as OPTION lines
			while ($content = sqlFetchArray($result)) {
				// Is extra set?
				if (!isset($content['extra'])) {
					// Set it to empty
					$content['extra'] = '';
				} // END - if

				$ret .= '<option value="' . $content['key'] . '"';

				if ($default == $content['key']) {
					// Selected by default
					$ret .= ' selected="selected"';
				} elseif (isset($disabled[$content['key']])) {
					// Disabled!
					$ret .= ' disabled="disabled"';
				}

				// Add it, if set
				if (!empty($content['extra'])) {
					$content['extra'] = ' (' . $content['extra'] . ')';
				} // END - if

				// Is the call-back function set?
				if (!empty($callback)) {
					// Call it
					$content['value'] = call_user_func_array($callback, array($content['value']));
				} // END - if

				// Finish option list
				$ret .= '>' . $content['value'] . $content['extra'] . '</option>';
			} // END - while
		} else {
			// No data found
			$ret = '<option value="x">{--SELECT_NONE--}</option>';
		}

		// Free memory
		sqlFreeResult($result);
	}

	// Return - hopefully - the requested data
	return $ret;
}

// Deletes a user account with given reason
function deleteUserAccount ($userid, $reason) {
	// Init points
	$data['points'] = '0';

	// Search for the points and user data
	$result = sqlQueryEscaped("SELECT
	(SUM(`p`.`points`) - `d`.`used_points`) AS `points`
FROM
	`{?_MYSQL_PREFIX?}_user_points` AS `p`
LEFT JOIN
	`{?_MYSQL_PREFIX?}_user_data` AS `d`
ON
	`p`.`userid`=`d`.`userid`
WHERE
	`p`.`userid`=%s
LIMIT 1",
		array(bigintval($userid)), __FUNCTION__, __LINE__);

	// Is there an entry?
	if (sqlNumRows($result) == 1) {
		// Save his points to add them to the jackpot
		$data = sqlFetchArray($result);

		// Delete points entries as well
		// @TODO Rewrite these lines to a filter
		sqlQueryEscaped("DELETE LOW_PRIORITY FROM `{?_MYSQL_PREFIX?}_user_points` WHERE `userid`=%s",
			array(bigintval($userid)), __FUNCTION__, __LINE__);

		// Update mediadata as well
		if (isExtensionInstalledAndNewer('mediadata', '0.0.4')) {
			// Update database
			updateMediadataEntry(array('total_points'), 'sub', $data['points']);
		} // END - if

		// Now, when we have all his points adds them do the jackpot!
		if (isExtensionActive('jackpot')) {
			addPointsToJackpot($data['points']);
		} // END - if
	} // END - if

	// Free the result
	sqlFreeResult($result);

	// Delete category selections as well...
	sqlQueryEscaped("DELETE LOW_PRIORITY FROM `{?_MYSQL_PREFIX?}_user_cats` WHERE `userid`=%s",
		array(bigintval($userid)), __FUNCTION__, __LINE__);

	// Remove from rallye if found
	// @TODO Rewrite this to a filter
	if (isExtensionActive('rallye')) {
		sqlQueryEscaped("DELETE LOW_PRIORITY FROM `{?_MYSQL_PREFIX?}_rallye_users` WHERE `userid`=%s",
			array(bigintval($userid)), __FUNCTION__, __LINE__);
	} // END - if

	// Add reason and translate points
	$data['text'] = $reason;

	// Now a mail to the user and that's all...
	$message = loadEmailTemplate('member_user_deleted', $data, $userid);
	sendEmail($userid, '{--ADMIN_DELETE_ACCOUNT--}', $message);

	// Ok, delete the account!
	sqlQueryEscaped("DELETE LOW_PRIORITY FROM `{?_MYSQL_PREFIX?}_user_data` WHERE `userid`=%s LIMIT 1", array(bigintval($userid)), __FUNCTION__, __LINE__);
}

// Gets the matching what name from module
function getWhatFromModule ($modCheck) {
	// Is the request element set?
	if (isGetRequestElementSet('what')) {
		// Then return this!
		return getRequestElement('what');
	} // END - if

	// Default is empty
	$what = '';

	// Check on given module
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'modCheck=' . $modCheck);
	switch ($modCheck) {
		case 'index': // Guest area
			// Is ext-sql_patches installed and newer than 0.0.5?
			if (isExtensionInstalledAndNewer('sql_patches', '0.0.5')) {
				// Use it from config
				$what = getIndexHome();
			} else {
				// Use default 'welcome'
				$what = 'welcome';
			}
			break;

		default: // Default for all other menus (getIndexHome() is for index module only)
			$what = 'welcome';
			break;
	} // END - switch

	// Return what value
	return $what;
}

// Returns HTML code with an option list of all categories
function generateCategoryOptionsList ($mode, $userid = NULL) {
	// Prepare WHERE statement
	$whereStatement = " WHERE `visible`='Y'";
	if (isAdmin()) $whereStatement = '';

	// Initialize array...
	$categories = array(
		'id'      => array(),
		'name'    => array(),
		'userids' => array()
	);

	// Get categories
	$result = sqlQuery('SELECT
	`id`,
	`cat`
FROM
	`{?_MYSQL_PREFIX?}_cats`
' . $whereStatement . '
ORDER BY
	`sort` ASC',
		__FUNCTION__, __LINE__);

	// Are there entries?
	if (!ifSqlHasZeroNums($result)) {
		// ... and begin loading stuff
		while ($content = sqlFetchArray($result)) {
			// Transfer some data
			$categories['id'][]   = $content['id'];
			array_push($categories['name'], $content['cat']);

			// Check which users are in this category
			$result_userids = sqlQueryEscaped("SELECT `userid` FROM `{?_MYSQL_PREFIX?}_user_cats` WHERE `cat_id`=%s AND `userid` != %s ORDER BY `userid` ASC",
				array(
					bigintval($content['id']),
					convertNullToZero($userid)
				), __FUNCTION__, __LINE__);

			// Init count
			$userid_cnt = '0';

			// Start adding all
			while ($data = sqlFetchArray($result_userids)) {
				// Add user count
				$userid_cnt += countSumTotalData($data['userid'], 'user_data', 'userid', 'userid', TRUE, runFilterChain('user_exclusion_sql', " AND `status`='CONFIRMED' AND `receive_mails` > 0"));
			} // END - while

			// Free memory
			sqlFreeResult($result_userids);

			// Add counter
			array_push($categories['userids'], $userid_cnt);
		} // END - while

		// Free memory
		sqlFreeResult($result);

		// Generate options
		$OUT = '';
		foreach ($categories['id'] as $key => $value) {
			$OUT .= '      <option value="' . $value . '">' . $categories['name'][$key] . ' (' . $categories['userids'][$key] . ' {--USERS_IN_CATEGORY--})</option>';
		} // END - foreach
	} else {
		// No cateogries are defined yet
		$OUT = '<option class="bad">{--MEMBER_NO_CATEGORIES--}</option>';
	}

	// Return HTML code
	return $OUT;
}

// Add bonus mail to queue
function addBonusMailToQueue ($subject, $text, $receiverList, $points, $seconds, $url, $categoryId, $mode='normal', $receiver=0) {
	// Is admin or bonus extension there?
	if (!isAdmin()) {
		// Abort here
		return FALSE;
	} elseif (!isExtensionActive('bonus')) {
		// Abort here
		return FALSE;
	}

	// Calculcate target sent
	$target = countSelection(explode(';', $receiverList));

	// Receiver is zero?
	if ($receiver == '0') {
		// Then auto-fix it
		$receiver = $target;
	} // END - if

	// HTML extension active?
	if (isExtensionActive('html_mail')) {
		// Add HTML mail
		sqlQueryEscaped("INSERT INTO `{?_MYSQL_PREFIX?}_bonus` (
	`subject`,
	`text`,
	`receivers`,
	`points`,
	`time`,
	`data_type`,
	`timestamp`,
	`url`,
	`cat_id`,
	`target_send`,
	`mails_sent`,
	`html_msg`
) VALUES (
	'%s',
	'%s',
	'%s',
	%s,
	%s,
	'NEW',
	UNIX_TIMESTAMP(),
	'%s',
	%s,
	%s,
	%s,
	'%s'
)",
		array(
			$subject,
			$text,
			$receiverList,
			$points,
			bigintval($seconds),
			$url,
			bigintval($categoryId),
			$target,
			bigintval($receiver),
			convertBooleanToYesNo($mode == 'html')
		), __FUNCTION__, __LINE__);
	} else {
		// Add regular mail
		sqlQueryEscaped("INSERT INTO `{?_MYSQL_PREFIX?}_bonus` (
	`subject`,
	`text`,
	`receivers`,
	`points`,
	`time`,
	`data_type`,
	`timestamp`,
	`url`,
	`cat_id`,
	`target_send`,
	`mails_sent`
) VALUES (
	'%s',
	'%s',
	'%s',
	%s,
	%s,
	'NEW',
	UNIX_TIMESTAMP(),
	'%s',
	%s,
	%s,
	%s
)",
		array(
			$subject,
			$text,
			$receiverList,
			$points,
			bigintval($seconds),
			$url,
			bigintval($categoryId),
			$target,
			bigintval($receiver),
		), __FUNCTION__, __LINE__);
	}
}

// Generate a receiver list for given category and maximum receivers
function generateReceiverList ($categoryId, $receiver, $mode = '') {
	// Init variables
	$extraColumns = '';
	$receiverList = '';
	$result       = FALSE;

	// Secure data
	$categoryId = bigintval($categoryId);
	$receiver   = bigintval($receiver);

	// Is the receiver zero and mode set?
	if (($receiver == '0') && (!empty($mode))) {
		// Auto-fix receiver maximum
		$receiver = getTotalReceivers($mode);
	} // END - if

	// Exclude (maybe exclude) testers
	$addWhere = runFilterChain('user_exclusion_sql', ' ');

	// Category given?
	if (isValidId($categoryId)) {
		// Select category
		$extraColumns  = "LEFT JOIN `{?_MYSQL_PREFIX?}_user_cats` AS `c` ON `d`.`userid`=`c`.`userid`";
		$addWhere = sprintf(' AND `c`.`cat_id`=%s', $categoryId);
	} // END - if

	// Exclude users in holiday?
	if (isExtensionInstalledAndNewer('holiday', '0.1.3')) {
		// Add something for the holiday extension
		$addWhere .= " AND `d`.`holiday_active`='N'";
	} // END - if

	// Run query
	$result = sqlQueryEscaped("SELECT
	`d`.`userid`
FROM
	`{?_MYSQL_PREFIX?}_user_data` AS `d`
	" . $extraColumns . "
WHERE
	`d`.`status`='CONFIRMED'
	" . $addWhere . "
ORDER BY
	`d`.`{?order_select?}` {?order_mode?}
LIMIT %s",
		array(
			$receiver
		), __FUNCTION__, __LINE__);

	// Entries found?
	if ((sqlNumRows($result) >= $receiver) && ($receiver > 0)) {
		// Load all entries
		while ($content = sqlFetchArray($result)) {
			// Add receiver when not empty
			if (!empty($content['userid'])) {
				$receiverList .= $content['userid'] . ';';
			} // END - if
		} // END - while

		// Free memory
		sqlFreeResult($result);

		// Remove trailing semicolon
		$receiverList = substr($receiverList, 0, -1);
	} // END - if

	// Return list
	return $receiverList;
}

// Recuce the amount of received emails for the receipients for given email
function reduceRecipientReceivedMails ($column, $id, $count) {
	// Search for mail in database
	$result = sqlQueryEscaped("SELECT * FROM `{?_MYSQL_PREFIX?}_user_links` WHERE `%s`=%s ORDER BY `userid` ASC LIMIT %s",
		array(
			$column,
			bigintval($id),
			$count
		), __FUNCTION__, __LINE__
	);

	// Are there entries?
	if (!ifSqlHasZeroNums($result)) {
		// Now load all userids for one big query!
		$userids = array();
		while ($data = sqlFetchArray($result)) {
			// By default reduce and found no emails
			$num = 0;

			// We must now look if he has already confirmed this mail, so might sound double, but it may resolve problems
			// @TODO Rewrite this to a filter
			if ((isset($data['stats_id'])) && (isValidId($data['stats_id']))) {
				// User email
				$num = countSumTotalData($data['userid'], 'user_stats_data', 'id', 'userid', TRUE, sprintf(" AND `stats_type`='mailid' AND `stats_data`=%s", bigintval($data['stats_id'])));
			} elseif ((isset($data['bonus_id'])) && (isValidId($data['bonus_id']))) {
				// Bonus mail
				$num = countSumTotalData($data['userid'], 'user_stats_data', 'id', 'userid', TRUE, sprintf(" AND `stats_type`='bonusid' AND `stats_data`=%s", bigintval($data['bonus_id'])));
			}

			// Reduce this users total received emails?
			if ($num === 0) {
				$userids[$data['userid']] = $data['userid'];
			} // END - if
		} // END - while

		if (isFilledArray($userids)) {
			// Now update all user accounts
			sqlQueryEscaped("UPDATE `{?_MYSQL_PREFIX?}_user_data` SET `emails_received`=`emails_received`-1 WHERE `userid` IN (%s) LIMIT %s",
				array(
					implode(',', $userids),
					count($userids)
				), __FUNCTION__, __LINE__);
		} else {
			// Nothing deleted
			displayMessage('{%message,ADMIN_MAIL_NOTHING_DELETED=' . $id . '%}');
		}
	} // END - if

	// Free result
	sqlFreeResult($result);
}

// Creates a new task
function createNewTask ($subject, $notes, $taskType, $userid = NULL, $adminId = NULL, $strip = TRUE) {
	// Insert the task data into the database
	sqlQueryEscaped("INSERT INTO `{?_MYSQL_PREFIX?}_task_system` (`assigned_admin`, `userid`, `status`, `task_type`, `subject`, `text`, `task_created`) VALUES (%s, %s, 'NEW', '%s', '%s', '%s', UNIX_TIMESTAMP())",
		array(
			convertZeroToNull($adminId),
			convertZeroToNull($userid),
			$taskType,
			$subject,
			$notes
		), __FUNCTION__, __LINE__, TRUE, $strip);

	// Return insert id which is the task id
	return getSqlInsertId();
}

// Updates last module / online time
function updateLastActivity ($userid) {
	// Is 'what' set?
	if (isWhatSet()) {
		// Run the update query
		sqlQueryEscaped("UPDATE
	`{?_MYSQL_PREFIX?}_user_data`
SET
	`{%%pipe,getUserLastWhatName%%}`='{%%pipe,getWhat%%}',
	`last_online`=UNIX_TIMESTAMP(),
	`REMOTE_ADDR`='{%%pipe,detectRemoteAddr%%}'
WHERE
	`userid`=%s
LIMIT 1",
		array(
			bigintval($userid)
		), __FUNCTION__, __LINE__);
	} else {
		// No what set, needs to be ignored (last_module is last_what)
		sqlQueryEscaped("UPDATE
	`{?_MYSQL_PREFIX?}_user_data`
SET
	`{%%pipe,getUserLastWhatName%%}`=NULL,
	`last_online`=UNIX_TIMESTAMP(),
	`REMOTE_ADDR`='{%%pipe,detectRemoteAddr%%}'
WHERE
	`userid`=%s
LIMIT 1",
		array(
			bigintval($userid)
		), __FUNCTION__, __LINE__);
	}
}

// List all given joined rows (callback function from XML)
function doGenericJoinedListEntries ($tableTemplate, $rowTemplate, $noEntryMessageId, $tableName, $tableJoinType, $tableJoinName, $joinOnLeftTable, $joinOnCondition, $joinOnRightTable, $columns, $whereColumns, $orderByColumns, $callbackColumns, $extraParameters = array(), $conditions = array(), $content = array()) {
	// Verify that tableName and columns are not empty
	if ((!is_array($tableName)) || (count($tableName) != 1)) {
		// No tableName specified
		reportBug(__FUNCTION__, __LINE__, 'tableName is not given. Please fix your XML,tableName[]=' . gettype($tableName) . '!=array,tableTemplate=' . $tableTemplate . ',rowTemplate=' . $rowTemplate);
	} elseif (!isFilledArray($columns)) {
		// No columns specified
		reportBug(__FUNCTION__, __LINE__, 'columns is not given. Please fix your XML,tableTemplate=' . $tableTemplate . ',rowTemplate=' . $rowTemplate . ',tableName[0]=' . $tableName[0]);
	}

	// This is the minimum query, so at least columns and tableName must have entries
	$sql = 'SELECT ';

	// Get the sql part back from given array
	$sql .= getSqlPartFromXmlArray($columns);

	// Add "FROM"
	$sql .= getSqlXmlFromTable($tableName);

	// Add "JOIN"
	$sql .= getSqlXmlJoinedTable($tableJoinType, $tableJoinName, $joinOnLeftTable, $joinOnCondition, $joinOnRightTable);

	// Add "WHERE"
	$sql .= getSqlXmlWhereConditions($whereColumns, $conditions);

	// Add "ORDER BY"
	$sql .= getSqlXmlOrderBy($orderByColumns);

	// Now handle all over to the inner function which will execute the listing
	doListEntries($sql, $tableTemplate, $noEntryMessageId, $rowTemplate, $callbackColumns, $extraParameters, $content);
}

// List all given rows (callback function from XML)
function doGenericListEntries ($tableTemplate, $rowTemplate, $noEntryMessageId, $tableName, $columns, $whereColumns, $orderByColumns, $callbackColumns, $extraParameters = array(), $conditions = array(), $content = array()) {
	// Verify that tableName and columns are not empty
	if ((!is_array($tableName)) || (count($tableName) != 1)) {
		// No tableName specified
		reportBug(__FUNCTION__, __LINE__, 'tableName is not given. Please fix your XML,tableName[]=' . gettype($tableName) . '!=array,tableTemplate=' . $tableTemplate . ',rowTemplate=' . $rowTemplate);
	} elseif (!isFilledArray($columns)) {
		// No columns specified
		reportBug(__FUNCTION__, __LINE__, 'columns is not given. Please fix your XML,tableTemplate=' . $tableTemplate . ',rowTemplate=' . $rowTemplate . ',tableName[0]=' . $tableName[0]);
	}

	// This is the minimum query, so at least columns and tableName must have entries
	$sql = 'SELECT ';

	// Get the sql part back from given array
	$sql .= getSqlPartFromXmlArray($columns);

	// Add "FROM"
	$sql .= getSqlXmlFromTable($tableName);

	// Add "WHERE"
	$sql .= getSqlXmlWhereConditions($whereColumns, $conditions);

	// Add "ORDER BY"
	$sql .= getSqlXmlOrderBy($orderByColumns);

	// Now handle all over to the inner function which will execute the listing
	doListEntries($sql, $tableTemplate, $noEntryMessageId, $rowTemplate, $callbackColumns, $extraParameters, $content);
}

// Do the listing of entries
function doListEntries ($sql, $tableTemplate, $noEntryMessageId, $rowTemplate, $callbackColumns, $extraParameters = array(), $content = array()) {
	// Run the SQL query
	$result = sqlQuery($sql, __FUNCTION__, __LINE__);

	// Are there some URLs left?
	if (!ifSqlHasZeroNums($result)) {
		// List all URLs
		$OUT = '';
		while ($row = sqlFetchArray($result)) {
			// "Translate" content
			foreach ($callbackColumns as $columnName => $callbackName) {
				// Fill the callback arguments
				$args = array($row[$columnName]);

				// Is there more to add?
				if (isset($extraParameters[$columnName])) {
					// Add them as well
					$args = merge_array($args, $extraParameters[$columnName]);
				} // END - if

				// Call the callback-function
				//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'callbackFunction=' . $callbackName . ',args=<pre>'.print_r($args, TRUE).'</pre>');
				// @TODO If the EL sub-system can support more than one parameter, this call_user_func_array() can be avoided
				$row[$columnName] = call_user_func_array($callbackName, $args);
			} // END - foreach

			// Load row template
			$OUT .= loadTemplate(trim($rowTemplate[0]), TRUE, $row);
		} // END - while

		// Is at least one entry set in content?
		if (isFilledArray($content)) {
			// Then add generic 'rows' element
			$content['rows'] = $OUT;
		} else {
			// Direct output is content
			$content = $OUT;
		}

		// Load main template
		loadTemplate(trim($tableTemplate[0]), FALSE, $content);
	} else {
		// No URLs in surfbar
		displayMessage('{--' .$noEntryMessageId[0] . '--}');
	}

	// Free result
	sqlFreeResult($result);
}

// Adds a given entry to the database
function doGenericAddEntries ($tableName, $columns = array(), $filterFunctions = array(), $extraValues = array(), $timeColumns = array(), $columnIndex = NULL) {
	//* DEBUG: */ die(__FUNCTION__.':columns=<pre>'.print_r($columns,TRUE).'</pre>,filterFunctions=<pre>'.print_r($filterFunctions,TRUE).'</pre>,extraValues=<pre>'.print_r($extraValues,TRUE).'</pre>,timeColumns=<pre>'.print_r($timeColumns,TRUE).'</pre>,columnIndex=<pre>'.print_r($columnIndex,TRUE).'</pre>,POST=<pre>'.print_r($_POST,TRUE).'</pre>');
	// Verify that tableName and columns are not empty
	if ((!is_array($tableName)) || (count($tableName) != 1)) {
		// No tableName specified
		reportBug(__FUNCTION__, __LINE__, 'tableName is not given. Please fix your XML,tableName[]=' . gettype($tableName) . '!=array: userIdColumn=' . $userIdColumn);
	} elseif (!isFilledArray($columns)) {
		// No columns specified
		reportBug(__FUNCTION__, __LINE__, 'columns is not given. Please fix your XML.');
	}

	// Init columns and value elements
	$sqlColumns = array();
	$sqlValues  = array();

	// Default is that all went fine
	$GLOBALS['__XML_PARSE_RESULT'] = TRUE;

	// Is there "time columns"?
	if (isFilledArray($timeColumns)) {
		// Then "walk" through all entries
		foreach ($timeColumns as $column) {
			// Convert all (possible) selections
			convertSelectionsToEpocheTimeInPostData($column . '_ye');
		} // END - foreach
	} // END - if

	// Add columns and values
	foreach ($columns as $key => $columnName) {
		//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'key=' . $key . ',columnName=' . $columnName);
		// Is columnIndex set?
		if (!is_null($columnIndex)) {
			// Check conditions
			//* DEBUG: */ die('columnName=<pre>'.print_r($columnName,TRUE).'</pre>columnIndex=<pre>'.print_r($columnIndex,TRUE).'</pre>'.debug_get_printable_backtrace());
			assert((is_array($columnName)) && (is_string($columnIndex)) && (isset($columnName[$columnIndex])));

			// Then use that index "blindly"
			$columnName = $columnName[$columnIndex];
		} // END - if

		// Debug message
		//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'key=' . $key . ',columnName[' . gettype($columnName) . ']=' . $columnName . ',filterFunctions=' . $filterFunctions[$key] . ',extraValues=' . intval(isset($extraValues[$key])) . ',extraValuesName=' . intval(isset($extraValues[$columnName . '_list'])));

		// Get value back (no array supported)
		$value = postRequestElement($columnName);

		// Is this an array and element 0 is set?
		if ((is_array($value)) && (isset($value[0]))) {
			// Then only take this
			$value = $value[0];
		} // END - if

		// Copy entry securely to the final arrays
		$sqlColumns[$key] = sqlEscapeString($columnName);
		$sqlValues[$key]  = sqlEscapeString($value);

		// Search for it
		$search = key(search_array($columns, 'column', $columnName));

		// Try to handle call-back functions and/or extra values on the list
		//* DEBUG: */ outputHtml($key . '/' . $columnName . '=<pre>'.print_r($columns,true).'</pre>search_array()=<pre>'.print_r(search_array($columns, 'column', $columnName), TRUE).'</pre>');
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'sqlValues[' . $key . '(' . gettype($key) . ')][' . gettype($sqlValues[$key]) . ']=' . $sqlValues[$key] . ' - BEFORE!');
		if (is_string($key)) {
			// Key is a string
			$sqlValues[$key] = doHandleExtraValues($filterFunctions, $extraValues, $key . '_list', $sqlValues[$key], array(''), $search);
		} else {
			// Is a number
			$sqlValues[$key] = doHandleExtraValues($filterFunctions, $extraValues, $key, $sqlValues[$key], array(''), $search);
		}

		// Is the value not a number?
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'sqlValues[' . $key . '(' . gettype($key) . ')][' . gettype($sqlValues[$key]) . ']=' . $sqlValues[$key] . ' - AFTER!');
		if (($sqlValues[$key] != 'NULL') && (is_string($sqlValues[$key]))) {
			// Add quotes around it
			$sqlValues[$key] = chr(39) . $sqlValues[$key] . chr(39);
		} // END - if

		// Is the value false?
		if ($sqlValues[$key] === FALSE) {
			// One "parser" didn't like it
			$GLOBALS['__XML_PARSE_RESULT'] = FALSE;
			break;
		} // END - if
	} // END - foreach

	// If all values are okay, continue
	if ($sqlValues[$key] !== FALSE) {
		// Build the SQL query
		$sql = 'INSERT INTO `{?_MYSQL_PREFIX?}_' . $tableName[0] . '` (`' . implode('`, `', $sqlColumns) . "`) VALUES (" . implode(',', $sqlValues) . ')';

		// Run the SQL query
		sqlQuery($sql, __FUNCTION__, __LINE__);

		// Add id number
		setPostRequestElement('id', getSqlInsertId());

		// Prepare filter data array
		$filterData = array(
			'mode'          => 'add',
			'table_name'    => $tableName,
			'content'       => postRequestArray(),
			'id'            => getSqlInsertId(),
			'subject'       => '',
			// @TODO Used generic 'userid' here
			'userid_column' => array('userid'),
			'raw_userid'    => array('userid'),
			'affected'      => sqlAffectedRows(),
			'sql'           => $sql,
		);

		// Send "build mail" out
		runFilterChain('send_build_mail', $filterData);
	} // END - if
}

// Edit rows by given id numbers
function doGenericEditEntriesConfirm ($tableName, $columns = array(), $filterFunctions = array(), $extraValues = array(), $timeColumns = array(), $editNow = array(FALSE), $idColumn = array('id'), $userIdColumn = array('userid'), $rawUserId = array('userid'), $cacheFiles = array(), $subject = '') {
	// Is there "time columns"?
	if (isFilledArray($timeColumns)) {
		// Then "walk" through all entries
		foreach ($timeColumns as $column) {
			// Convert all (possible) selections
			convertSelectionsToEpocheTimeInPostData($column . '_ye');
		} // END - foreach
	} // END - if

	// Change them all
	$affected = '0';
	foreach (postRequestElement($idColumn[0]) as $id => $sel) {
		// Secure id number
		$id = bigintval($id);

		// Prepare content array (new values)
		$content = array();

		// Prepare SQL for this row
		$sql = sprintf('UPDATE `{?_MYSQL_PREFIX?}_%s` SET',
			sqlEscapeString($tableName[0])
		);

		// "Walk" through all entries
		foreach (postRequestArray() as $key => $entries) {
			// Skip raw userid which is always invalid
			if (($key == $rawUserId[0]) || ($key == ($rawUserId[0] . '_raw')) || ($key == 'do_edit')) {
				// Continue with next field
				//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'key=' . $key . ',idColumn[0]=' . $idColumn[0] . ',rawUserId=' . $rawUserId[0]);
				continue;
			} // END - if

			// Debug message
			//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'key=' . $key . ',id=' . $id . ',idColumn[0]=' . $idColumn[0] . ',entries=<pre>'.print_r($entries,TRUE).'</pre>');

			// Is entries an array?
			if (($key != $idColumn[0]) && (is_array($entries)) && (isset($entries[$id]))) {
				// Search for the right array index
				$search = key(search_array($columns, 'column', $key));

				// Add this entry to content
				$content[$key] = $entries[$id];

				// Debug message
				//* BUG: */ die($key.'/'.$id.'/'.$search.'=<pre>'.print_r($columns,TRUE).'</pre><pre>'.print_r($filterFunctions,TRUE).'</pre>');

				// Handle possible call-back functions and/or extra values
				$entries[$id] = doHandleExtraValues($filterFunctions, $extraValues, $key, $entries[$id], $userIdColumn, $search);

				// Add key/value pair to SQL string
				$sql .= addKeyValueSql($key, $entries[$id]);
			} elseif (($key != $idColumn[0]) && (!is_array($entries))) {
				// Search for it
				$search = key(search_array($columns, 'column', $key));
				//* BUG: */ die($key.'/<pre>'.print_r($search, TRUE).'</pre>=<pre>'.print_r($columns, TRUE).'</pre>');

				// Debug message
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'key=' . $key . ',entries[' . gettype($entries) . ']=' . $entries . ',search=' . $search . ' - BEFORE!');

				// Add normal entries as well
				$content[$key] = $entries;

				// Handle possible call-back functions and/or extra values
				$entries = doHandleExtraValues($filterFunctions, $extraValues, $key, $entries, $userIdColumn, $search);

				// Debug message
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'key=' . $key . ',entries[' . gettype($entries) . ']=' . $entries . ',search=' . $search . ' - AFTER!');

				// Add key/value pair to SQL string
				$sql .= addKeyValueSql($key, $entries);
			}
		} // END - foreach

		// Finish SQL command
		$sql = substr($sql, 0, -1) . " WHERE `" . sqlEscapeString($idColumn[0]) . "`=" . $id;
		if ((isset($rawUserId[0])) && (isset($userIdColumn[0])) && (isPostRequestElementSet($rawUserId[0])) && (!is_array(postRequestElement($rawUserId[0])))) {
			// Add user id as well
			$sql .= ' AND `' . $userIdColumn[0] . '`=' . bigintval(postRequestElement($rawUserId[0]));
		} // END - if
		$sql .= " LIMIT 1";

		// Run this query
		//* BUG: */ die($sql.'<pre>'.print_r(postRequestArray(), TRUE).'</pre>');
		sqlQuery($sql, __FUNCTION__, __LINE__);

		// Add affected rows
		$edited = sqlAffectedRows();
		$affected += $edited;

		// Load all data from that id
		$result = sqlQueryEscaped("SELECT * FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s`=%s LIMIT 1",
			array(
				$tableName[0],
				$idColumn[0],
				$id
			), __FUNCTION__, __LINE__);

		// Fetch the data and merge it into $content
		$content = merge_array($content, sqlFetchArray($result));

		// Prepare filter data array
		$filterData = array(
			'mode'          => 'edit',
			'table_name'    => $tableName,
			'content'       => $content,
			'id'            => $id,
			'subject'       => $subject,
			'userid_column' => $userIdColumn,
			'raw_userid'    => $rawUserId,
			'affected'      => $edited,
			'sql'           => $sql,
		);

		// Send "build mail" out
		runFilterChain('send_build_mail', $filterData);

		// Free the result
		sqlFreeResult($result);
	} // END - foreach

	// Delete cache?
	if ((isFilledArray($cacheFiles)) && (!empty($cacheFiles[0]))) {
		// Delete cache file(s)
		foreach ($cacheFiles as $cache) {
			// Skip any empty entries
			if (empty($cache)) {
				// This may cause trouble in loadCacheFile()
				continue;
			} // END - if

			// Use rebuildCache() to delete it
			rebuildCache($cache);
		} // END - foreach
	} // END - if

	// Return affected rows
	return $affected;
}

// Delete rows by given id numbers
function doGenericDeleteEntriesConfirm ($tableName, $columns = array(), $filterFunctions = array(), $extraValues = array(), $deleteNow = array(FALSE), $idColumn = array('id'), $userIdColumn = array('userid'), $rawUserId = array('userid'), $cacheFiles = array()) {
	// The base SQL command:
	$sql = "DELETE LOW_PRIORITY FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s` IN (%s)";

	// Is a user id provided?
	//* BUG: */ die('<pre>'.print_r($rawUserId,TRUE).'</pre><pre>'.print_r($userIdColumn,TRUE).'</pre>');
	if ((isset($rawUserId[0])) && (isset($userIdColumn[0])) && (isPostRequestElementSet($rawUserId[0])) && (!is_array(postRequestElement($rawUserId[0])))) {
		// Add user id as well
		$sql .= ' AND `' . $userIdColumn[0] . '`=' . bigintval(postRequestElement($rawUserId[0]));
	} // END - if

	// $idColumn[0] in POST must be an array again
	if (!is_array(postRequestElement($idColumn[0]))) {
		// This indicates that you have conflicting form field naming with XML names
		reportBug(__FUNCTION__, __LINE__, 'You have a wrong form field element, idColumn[0]=' . $idColumn[0]);
	} // END - if

	// Delete them all
	//* BUG: */ die($sql.'<pre>'.print_r($tableName,TRUE).'</pre><pre>'.print_r($columns,TRUE).'</pre><pre>'.print_r($filterFunctions,TRUE).'</pre><pre>'.print_r($extraValues,TRUE).'</pre><pre>'.print_r($deleteNow,TRUE).'</pre><pre>'.print_r($idColumn,TRUE).'</pre>');
	$idList = '';
	foreach (postRequestElement($idColumn[0]) as $id => $sel) {
		// Is id zero?
		if (!isValidId($id)) {
			// Then skip this
			continue;
		} // END - if

		// Is there a userid?
		if (isPostRequestElementSet($userIdColumn[0])) {
			// Load all data from that id
			$result = sqlQueryEscaped("SELECT * FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s`=%s LIMIT 1",
				array(
					$tableName[0],
					$idColumn[0],
					$id
				), __FUNCTION__, __LINE__);

			// Fetch the data
			$content = sqlFetchArray($result);

			// Free the result
			sqlFreeResult($result);

			// Send "build mails" out
			sendGenericBuildMails('delete', $tableName, $content, $id, '', $userIdColumn);
		} // END - if

		// Add id number
		$idList .= $id . ',';
	} // END - foreach

	// Run the query
	sqlQueryEscaped($sql,
		array(
			$tableName[0],
			$idColumn[0],
			convertNullToZero(substr($idList, 0, -1))
		), __FUNCTION__, __LINE__);

	// Return affected rows
	return sqlAffectedRows();
}

// Build a special template list
// @TODO cacheFiles is not yet supported
function doGenericListBuilder ($prefix, $listType, $tableName, $columns, $filterFunctions, $extraValues, $idColumn, $userIdColumn, $rawUserId = array('userid'), $content = array()) {
	// $tableName and $idColumn must bove be arrays!
	if ((!is_array($tableName)) || (count($tableName) != 1)) {
		// $tableName is no array
		reportBug(__FUNCTION__, __LINE__, 'tableName[]=' . gettype($tableName) . '!=array: userIdColumn=' . $userIdColumn);
	} elseif (!is_array($idColumn)) {
		// $idColumn is no array
		reportBug(__FUNCTION__, __LINE__, 'idColumn[]=' . gettype($idColumn) . '!=array: userIdColumn=' . $userIdColumn);
	} elseif ((!is_array($userIdColumn)) || (count($userIdColumn) != 1)) {
		// $tableName is no array
		reportBug(__FUNCTION__, __LINE__, 'userIdColumn[]=' . gettype($userIdColumn) . '!=array: userIdColumn=' . $userIdColumn);
	}

	// Init row output
	$OUT = '';

	// "Walk" through all entries
	//* DEBUG: */ reportBug(__FUNCTION__, __LINE__, 'listType=<pre>'.print_r($listType,TRUE).'</pre>,tableName<pre>'.print_r($tableName,TRUE).'</pre>,columns=<pre>'.print_r($columns,TRUE).'</pre>,filterFunctions=<pre>'.print_r($filterFunctions,TRUE).'</pre>,extraValues=<pre>'.print_r($extraValues,TRUE).'</pre>,idColumn=<pre>'.print_r($idColumn,TRUE).'</pre>,userIdColumn=<pre>'.print_r($userIdColumn,TRUE).'</pre>,rawUserId=<pre>'.print_r($rawUserId,TRUE).'</pre>');
	foreach (postRequestElement($idColumn[0]) as $id => $selected) {
		// Secure id number
		$id = bigintval($id);

		// Get result from a given column array and table name
		$result = getSqlResultFromArray($tableName[0], $columns, $idColumn[0], $id, __FUNCTION__, __LINE__);

		// Is there one entry?
		if (sqlNumRows($result) == 1) {
			// Load all data
			$row = sqlFetchArray($result);

			// Filter all data
			foreach ($row as $key => $value) {
				// Search index
				$idx  = searchXmlArray($key, $columns, 'column');

				// Default name is NULL
				$name = NULL;

				// Is the name there?
				if (isset($columns[$idx]['name'])) {
					// Then use it
					$name = $columns[$idx]['name'];
				} // END - if

				// Skip any missing entries
				if ($idx === FALSE) {
					// Skip this one
					//* DEBUG: */ reportBug(__FUNCTION__, __LINE__, 'key=' . $key . ' - SKIPPED!');
					continue;
				} // END - if

				// Is there a userid?
				//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'key=' . $key . ',userIdColumn=' . $userIdColumn[0]);
				if ($key == $userIdColumn[0]) {
					// Add it again as raw id
					//* DEBUG: */ reportBug(__FUNCTION__, __LINE__, 'key=' . $key . ',userIdColumn=' . $userIdColumn[0]);
					$row[$userIdColumn[0]] = convertZeroToNull($value);
					$row[$userIdColumn[0] . '_raw'] = $row[$userIdColumn[0]];
				} // END - if

				// If the key matches the idColumn variable, we need to temporary remember it
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'key=' . $key . ',idColumn=' . $idColumn[0] . ',value=' . $value);
				if ($key == $idColumn[0]) {
					/*
					 * Found, so remember it securely (to make sure only id
					 * numbers can pass, don't use alpha-numerical values!)
					 */
					//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'value=' . $value . ' - set as ' . $prefix . '_list_builder_id_value!');
					$GLOBALS[$prefix . '_list_builder_id_value'] = bigintval($value);
				} // END - if

				// Try to handle call-back functions and/or extra values
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'idx=' . $idx . ',row[' . $key . ']=' . $row[$key]);
				//if ($key == 'forced_campaign_created') die($idx.'=<pre>'.print_r($columns,true).'</pre><pre>'.print_r($extraValues,true).'</pre>');
				$row[$key] = doHandleExtraValues($filterFunctions, $extraValues, $idx, $row[$key], $userIdColumn, ((!is_null($name)) ? $name : $key), $id);
			} // END - foreach

			// Then list it
			$OUT .= loadTemplate(sprintf('%s_%s_%s_row',
				$prefix,
				$listType,
				$tableName[0]
				), TRUE, $row
			);
		} // END - if

		// Free the result
		sqlFreeResult($result);
	} // END - foreach

	// Is there an entry in $content?
	if (isFilledArray($content)) {
		// Use generic 'rows'
		$content['rows'] = $OUT;
	} else {
		// Use direct output
		$content = $OUT;
	}

	// Load master template
	loadTemplate(
		sprintf('%s_%s_%s',
			$prefix,
			$listType,
			$tableName[0]
		), FALSE, $content
	);
}

// Adds key/value pair to a working SQL string together
function addKeyValueSql ($key, $value) {
	// Init SQL
	$sql = '';

	// Is it NULL?
	if (($value == 'NULL') || (is_null($value))) {
		// Add key with NULL
		$sql .= sprintf(' `%s`=NULL,',
			sqlEscapeString($key)
		);
	} elseif ((is_double($value)) || (is_float($value)) || (is_int($value))) {
		// Is a number, so addd it directly
		$sql .= sprintf(' `%s`=%s,',
			sqlEscapeString($key),
			$value
		);
	} else {
		// Else add the value escape'd
		$sql .= sprintf(" `%s`='%s',",
			sqlEscapeString($key),
			sqlEscapeString($value)
		);
	}

	// Return SQL string
	return $sql;
}

// "Getter" for an array from given table and columns
function getArrayFromTable ($tableName, $columns, $orderBy, $ordered = 'ASC', $whereSql = '') {
	// The table must exist
	assert(ifSqlTableExists($tableName));

	// Search for it
	$result = sqlQueryEscaped('SELECT
	`' . implode('`, `', $columns) . '`
FROM
	`{?_MYSQL_PREFIX?}_%s`
' . $whereSql . '
ORDER BY
	`%s` %S',
		array(
			$tableName,
			$orderBy,
			$ordered
		), __FUNCTION__, __LINE__
	);

	// Init array
	$rows = array();

	// Are there entries?
	if (!ifSqlHasZeroNums($result)) {
		// Load all entries
		while ($row = sqlFetchArray($result)) {
			array_push($rows, $row);
		} // END - while
	} // END - if

	// Free result
	sqlFreeResult($result);

	// Return all found rows
	return $rows;
}

// [EOF]
?>
